"""DAQ (Data AcQuisition) module for NI DAQ USB devices. Requires binary NIDAQserver."""

import os, time, numpy, sys

class NIDAQException (Exception):
    """Parent class for all exceptions thrown by the NIDAQ module."""
    def __str__(self):
        return self.__doc__

class NoNIDAQPresent(NIDAQException):
    """No NIDAQ is connected to the system."""

class NIDAQserverNotFound (NIDAQException):
    """Indicates a NIDAQserver binary could not be found in the current PATH nor in the current folder."""
    pass

class NIDAQServerException(NIDAQException):
    """Indicates the NIDAQserver was found but could not be used."""
    def __init__(self,error=""):
        self.error = error

    def __str__(self):
        return self.error

def initNIDAQ():
    """Initializes the NIDAQ and returns a procedure that can be used to acquire samples from it."""

    NIDAQserverfile = None

    isNIDAQconnected = False

    # Look for NIDAQserver binary
    for path in os.environ.get('PATH').split(':') + ['.']:
        if os.path.exists(os.path.join(path,'NIDAQserver')):
            NIDAQserverfile = os.path.join(path,'NIDAQserver')
            break

    # Check for running NIDAQserver process
    isNIDAQRunning = lambda : os.popen('ps -A').read().find('NIDAQserver') != -1

    if NIDAQserverfile is None and not isNIDAQRunning(): #if neither exists, throw exception
        raise NIDAQserverNotFound()

    if not isNIDAQRunning(): #if NIDAQserver is not running, start it with default settings
        for line in os.popen('/etc/natinst/nidaqmxbase/bin/daqmxbase_listdevices').read().split('\n'):
            if line.find('Dev') != -1:
                isNIDAQconnected = True
                break

        if not isNIDAQconnected:
            raise NoNIDAQPresent()
        
        pipe = os.popen(NIDAQserverfile + " --output &")
	sys.stdout.write("Starting NIDAQserver, this may take a little while...")
	pipe.read(400)
	sys.stdout.write('done\n')	
        if not isNIDAQRunning():
            raise NIDAQServerException("Could not start the NIDAQserver")

    settings = {}
    exec open('/tmp/NIDAQserver/log') in settings
    numberOfChannels = len(settings['channels'].split(','))
    samplesPerChannel = int(settings['samplesperchan'])

    def read(samples=1):
        """Return a channels * /samples/ numpy array containing the /samples/ latest readings from the NIDAQ
        (the most recent reading is at index 0). For example, to access the 5th latest value of the third analog
        channel (Ai2), type: read(5)[4][2]. For most applications, pulling a single sample is sufficient."""

        try:
            binaryStream = open("/tmp/NIDAQserver/binary2_module", "rb")
            chunk = binaryStream.read(samplesPerChannel * numberOfChannels * 8) # /numberOfChannels/ channels of 64 bits (8 bytes) doubles
            binaryStream.close() # We open and close the FIFO pipe each time to force the kernel to flush it, ensuring we have the latest data
        except IOError:
            raise NIDAQServerException('Could not acquire samples, make sure NIDAQ is plugged and call NIDAQ.initNIDAQ() again.')
        array = numpy.fromstring(chunk, numpy.float64, samplesPerChannel * numberOfChannels)
        array.shape = (array.size / numberOfChannels, numberOfChannels)
        return array[:samples]

    return read

def initNIDAQOutput():
    """Initializes the NIDAQ and returns a write function."""

    initNIDAQ() 
    import struct
    def write(sample):
        if hasattr(sample,'__float__'):
            sample = float(sample)
        assert type(sample) == float or type(sample) == int, "sample must be an integer or a floating point value!"
        assert sample >= -10.0 and sample <= 10.0, "sample must be between -10V and 10V!"
        try:
            toNIDAQ = open("/tmp/NIDAQserver/toNIDAQ","wb")
            toNIDAQ.write(struct.pack('d',sample))
            toNIDAQ.close()
        except IOError:
            pass # Rarely a sample can be dropped (IO Sync issue), need to implement a neater way around this
    return write

def medianFilter(readProcedure=initNIDAQ(),filterSize=8):
    """Takes a read procedure and a filterSize integer and returns a median-filtered
    read procedure."""

    if type(filterSize) != int:
        raise TypeError('filterSize has to be an integer!')
    if filterSize < 1:
        raise ValueError('filterSize has to be positive!')

    def median(array):
        # local procedure used to compute median

        array.sort(axis=0)
        
        if filterSize % 2 == 1: # Odd number, easy
            return array[filterSize // 2]
        else:
            return 0.5 * (array[filterSize // 2 - 1] + array[filterSize // 2])

    def readMedianFiltered(samples=1):
        """Returns 'samples' median filtered samples."""
        data = []
        for i in range(samples):
            data.append(median(readProcedure(filterSize)))
        return numpy.array(data)

    return readMedianFiltered
